import asyncio
import logging
from typing import Callable, Dict, Optional, Protocol, runtime_checkable

from pydantic import BaseModel, Field

from fluss.api.schema import (ArgNodeFragment, ArkitektNodeFragment,
                              FlowFragment, KwargNodeFragment,
                              ReactiveNodeFragment, ReturnNodeFragment,
                              RunMutationStart, aget_flow, arun, arunlog,
                              asnapshot, atrack)
from koil.types import Contextual
from reaktion.atoms.transport import AtomTransport
from reaktion.atoms.utils import atomify
from reaktion.contractors import (NodeContractor, arkicontractor,
                                  localcontractor)
from reaktion.events import EventType, InEvent, OutEvent
from reaktion.utils import connected_events
from rekuest.actors.base import Actor
from rekuest.actors.functional import AsyncFuncActor, AsyncGenActor
from rekuest.api.schema import (AssignationLogLevel, AssignationStatus,
                                NodeFragment, NodeKind, ProvisionFragment,
                                ProvisionStatus, ReservationFragment,
                                ReservationStatus, TemplateFragment, afind,
                                aget_template)
from rekuest.messages import Assignation, Provision
from rekuest.postmans.utils import (ReservationContract, RPCContract, arkiuse,
                                    localuse)

logger = logging.getLogger(__name__)


class NodeState(BaseModel):
    latestevent: OutEvent


class FlowActor(Actor):
    is_generator: bool = False
    flow: FlowFragment
    contracts: Dict[str, RPCContract] = Field(default_factory=dict)
    expand_inputs: bool = False
    shrink_outputs: bool = False
    provided = False
    is_generator: bool = False
    nodeContractor: NodeContractor = arkicontractor

    # Functionality for running the flow

    # Assign Related Functionality
    run_mutation: Callable = arun
    snapshot_mutation: Callable = asnapshot
    track_mutation: Callable = atrack

    atomifier: Callable = atomify
    """ Atomifier is a function that takes a node and returns an atom """

    run_states: Dict[
        str,
        Dict[str, NodeState],
    ] = Field(default_factory=dict)

    reservation_state: Dict[str, ReservationFragment] = Field(default_factory=dict)
    _lock = None

    async def on_provide(self, provision: Provision):
        self._lock = asyncio.Lock()

        argNode = [x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)][
            0
        ]

        kwargNode = [
            x for x in self.flow.graph.nodes if isinstance(x, KwargNodeFragment)
        ][0]

        returnNode = [
            x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)
        ][0]

        arkitektNodes = [
            x for x in self.flow.graph.nodes if isinstance(x, ArkitektNodeFragment)
        ]

        self.contracts = {
            node.id: await self.nodeContractor(node, self)
            for node in arkitektNodes
        }

        futures = [contract.aenter() for contract in self.contracts.values()]
        await asyncio.gather(*futures)

        self.provided = True
        await self.on_reservation_change(None)

    async def on_reservation_change(self, status: ReservationStatus):
        async with self._lock:
            unactive_reservations = [
                res
                for res in self.contracts.values()
                if res.state != ReservationStatus.ACTIVE
            ]
            if self.provided:
                if len(unactive_reservations) > 0:
                    await self.transport.change_provision(
                        self.provision.provision,
                        status=ProvisionStatus.CRITICAL,
                    )
                else:
                    await self.transport.change_provision(
                        self.provision.provision,
                        status=ProvisionStatus.ACTIVE,
                    )

    async def on_assign(self, assignation: Assignation):

        run = await self.run_mutation(
            assignation=assignation.assignation, flow=self.flow
        )

        await self.aass_log(assignation.assignation, "Starting")

        t = 0
        state = {}
        await self.snapshot_mutation(run=run, events=list(state.values()), t=t)

        try:
            event_queue = asyncio.Queue()

            atomtransport = AtomTransport(queue=event_queue)

            argNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ArgNodeFragment)
            ][0]
            kwargNode = [
                x for x in self.flow.graph.nodes if isinstance(x, KwargNodeFragment)
            ][0]
            returnNode = [
                x for x in self.flow.graph.nodes if isinstance(x, ReturnNodeFragment)
            ][0]

            participatingNodes = [
                x
                for x in self.flow.graph.nodes
                if isinstance(x, ArkitektNodeFragment)
                or isinstance(x, ReactiveNodeFragment)
            ]

            await self.aass_log(assignation.assignation, "Set up the graph")

            async def ass_log(assignation: Assignation, level, message):
                await self.aass_log(assignation.assignation, level, message)
                logging.info(f"{assignation}, {message}")

            atoms = {
                x.id: self.atomifier(
                    x, atomtransport, self.contracts, assignation, alog=ass_log
                )
                for x in participatingNodes
            }

            await self.aass_log(assignation.assignation, "Atomification complete")

            await asyncio.gather(*[atom.aenter() for atom in atoms.values()])

            tasks = [asyncio.create_task(atom.start()) for atom in atoms.values()]




            stream = argNode.outstream[0]
            value = [assignation.args[key] for key, index in enumerate(stream)]

            initial_event = OutEvent(
                handle=f"return_0",
                type=EventType.NEXT,
                source=argNode.id,
                value=value,
            )
            initial_done_event = OutEvent(
                handle=f"return_0",
                type=EventType.COMPLETE,
                source=argNode.id,
            )

            logger.info(f"Putting initial event {initial_event}")

            await event_queue.put(initial_event)
            await event_queue.put(initial_done_event)




            edge_targets = [e.target for e in self.flow.graph.edges]
            nodes_without_instream = [x for x in participatingNodes if len(x.instream[0]) == 0 and x.id not in edge_targets]


            logger.error(f"Nodes without instream: {nodes_without_instream}")
            for node in nodes_without_instream:

                assert node.id in atoms, "Atom not found. Should not happen."
                atom = atoms[node.id]

                initial_event = InEvent(
                        target=node.id,
                        handle="arg_0",
                        type=EventType.NEXT,
                        value=[],
                )
                done_event = InEvent(
                        target=node.id,
                        handle="arg_0",
                        type=EventType.COMPLETE,
                )

                await atom.put(initial_event)
                await atom.put(done_event)



            complete = False

            returns = []

            while not complete:
                event: OutEvent = await event_queue.get()
                event_queue.task_done()

                if self.flow.brittle:
                    if event.type == EventType.ERROR:
                        raise event.value

                track = await self.track_mutation(
                    run=run,
                    source=event.source,
                    handle=event.handle,
                    value=[str(i) for i in list(event.value)]
                    if event.value and not isinstance(event.value, Exception)
                    else event.value,
                    type=event.type,
                    t=t,
                )
                state[event.source] = track.id

                t += 1

                if t % 3 == 0:
                    await self.snapshot_mutation(
                        run=run, events=list(state.values()), t=t
                    )

                spawned_events = connected_events(self.flow.graph, event)
                if not spawned_events:
                    logger.warning(f"No events spawned from {event}")

                for spawned_event in spawned_events:
                    logger.info(f"-> {spawned_event}")

                    if spawned_event.target == returnNode.id:

                        if spawned_event.type == EventType.NEXT:
                            returns = spawned_event.value
                            if self.is_generator:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.YIELD,
                                    returns=returns,
                                )

                        if spawned_event.type == EventType.ERROR:
                            raise spawned_event.value

                        if spawned_event.type == EventType.COMPLETE:
                            complete = True
                            if not self.is_generator:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.RETURNED,
                                    returns=returns,
                                )
                            else:
                                await self.transport.change_assignation(
                                    assignation.assignation,
                                    status=AssignationStatus.DONE,
                                )

                    else:
                        assert (
                            spawned_event.target in atoms
                        ), "Unknown target. Your flow is connected wrong"
                        if spawned_event.target in atoms:
                            await atoms[spawned_event.target].put(spawned_event)

            for task in tasks:
                task.cancel()

            await asyncio.gather(*tasks, return_exceptions=True)

        except asyncio.CancelledError as e:

            for task in tasks:
                task.cancel()

            try:
                await asyncio.wait_for(
                    asyncio.gather(*tasks, return_exceptions=True), timeout=4
                )
            except asyncio.TimeoutError:
                pass

            await self.transport.change_assignation(
                assignation.assignation, status=AssignationStatus.CANCELLED
            )

        except Exception as e:
            logging.critical(f"Assignation {assignation} failed", exc_info=True)

            await self.aass_log(
                assignation.assignation, message=repr(e), level=AssignationStatus.ERROR
            )
            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.CRITICAL,
                message=repr(e),
            )

    async def on_unprovide(self):

        for contract in self.contracts.values():
            await contract.aexit()



import asyncio
from lib2to3.pytree import Base
from typing import Dict, Tuple
from reaktion.agent import ReaktionAgent
from rekuest.agents.errors import ProvisionException

from rekuest.agents.stateful import StatefulAgent
from rekuest.contrib.fakts.websocket_agent_transport import FaktsWebsocketAgentTransport
from rekuest.postmans.utils import arkiuse
from pydantic import Field
from rekuest.actors.base import Actor
from rekuest.agents.base import BaseAgent
from rekuest.api.schema import (
    PortInput,
    DefinitionInput,
    NodeFragment,
    NodeKind,
    TemplateFragment,
    acreate_template,
    adelete_node,
    afind,
    aget_provision,
    get_template,
)
from rekuest.messages import Assignation, Provision
from rekuest.postmans.utils import ReservationContract
from fakts.fakts import Fakts
from fluss.api.schema import (
    ArgNodeFragment,
    FlowFragment,
    KwargNodeFragment,
    ReturnNodeFragment,
)
from rekuest.widgets import SliderWidget, StringWidget
from arkitekt.apps.fluss import FlussApp
from arkitekt.apps.unlok import UnlokApp
from arkitekt.apps.rekuest import ArkitektRekuest, RekuestApp
from reaktion.utils import infer_kind_from_graph
from arkitekt.apps import Arkitekt
from fakts.grants.remote.device_code import DeviceCodeGrant
from fakts.fakts import Fakts
from fakts.grants import CacheGrant
from herre.grants import CacheGrant as HerreCacheGrant
from herre.grants.oauth2.refresh import RefreshGrant
from herre.grants.fakts import FaktsGrant
from fakts.discovery import StaticDiscovery
from herre import Herre
from arkitekt.apps.fluss import ConnectedFluss
import logging
from rich.logging import RichHandler


class ConnectedApp(FlussApp, RekuestApp, UnlokApp):
    pass


identifier = "fluss"
version = "latest"
url = "http://localhost:8000/f/"

app = ConnectedApp(
    fakts=Fakts(
        grant=CacheGrant(
            cache_file=f"{identifier}-{version}_cache.json",
            hash=f"{identifier}-{version}-{url}",
            grant=DeviceCodeGrant(
                identifier=identifier,
                version=version,
                discovery=StaticDiscovery(base_url=url),
            ),
        )
    ),
    herre=Herre(
        grant=HerreCacheGrant(
            cache_file=f"{identifier}-{version}_herre_cache.json",
            hash=f"{identifier}-{version}-{url}",
            grant=RefreshGrant(grant=FaktsGrant()),
        ),
    ),
    rekuest=ArkitektRekuest(
        agent=ReaktionAgent(
            transport=FaktsWebsocketAgentTransport(fakts_group="rekuest.agent")
        )
    ),
)


@app.rekuest.register(
    widgets={"description": StringWidget(as_paragraph=True)},
    interfaces=["fluss:deploy"],
)
async def deploy_graph(
    flow: FlowFragment,
    name: str = None,
    description: str = None,
) -> TemplateFragment:
    """Deploy Flow

    Deploys a Flow as a Template

    Args:
        graph (FlowFragment): The Flow
        name (str, optional): The name of this Incarnation
        description (str, optional): The name of this Incarnation

    Returns:
        TemplateFragment: The created template
    """
    assert flow.name, "Graph must have a Name in order to be deployed"

    print([x.dict(by_alias=True) for x in flow.graph.args])
    print([x.dict(by_alias=True) for x in flow.graph.returns])


    args = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.args]
    returns = [PortInput(**x.dict(by_alias=True)) for x in flow.graph.returns]

    template = await acreate_template(
        DefinitionInput(
            name=name or flow.workspace.name,
            interface=f"flow-{flow.id}",
            kind=infer_kind_from_graph(flow.graph),
            args=args,
            returns=returns,
            description=description,
            interfaces=["workflow", f"diagram:{flow.workspace.id}", f"flow:{flow.id}"],
        ),
        instance_id=app.rekuest.agent.instance_id,
        params={"flow": flow.id},
        extensions=["flow"],
    )

    return template


@app.rekuest.register()
async def undeploy_graph(
    flow: FlowFragment,
):
    """Undeploy Flow

    Undeploys graph, no user will be able to reserve this graph anymore

    Args:
        graph (FlowFragment): The Flow

    """
    assert flow.name, "Graph must have a Name in order to be deployed"

    x = await afind(interface=flow.hash)

    await adelete_node(x)
    return None


@app.rekuest.register(widgets={"interval": SliderWidget(min=0, max=100)})
async def timer(interval: int = 4) -> int:
    """Timer

    A simple timer that prints the current time every interval seconds

    Args:
        interval (int, optional): The interval in seconds. Defaults to 4.

    Returns:
        int: The current interval (iteration)

    """
    i = 0
    while True:
        i += 1
        yield i
        await asyncio.sleep(interval)


async def main():
    async with app:
        await app.rekuest.run()


if __name__ == "__main__":

    asyncio.run(main())



from asyncio.tasks import create_task
from reaktion.actor import FlowActor
from rekuest.agents.errors import ProvisionException
from rekuest.agents.stateful import StatefulAgent
import logging
from rekuest.actors.base import Actor
from fluss.api.schema import aget_flow
from rekuest.api.schema import aget_template, NodeKind
from rekuest.messages import Provision
from reaktion.errors import ReaktionError

logger = logging.getLogger(__name__)


class ReaktionAgent(StatefulAgent):
    async def aspawn_actor(self, prov: Provision) -> Actor:
        """Spawns an Actor from a Provision"""
        try:
            actor_builder = self._templateActorBuilderMap[prov.template]
            actor = actor_builder(provision=prov, transport=self.transport)
        except KeyError as e:
            try:
                x = await aget_template(prov.template)
                assert "flow" in x.params, "Template is not a flow"

                t = await aget_flow(id=x.params["flow"])
                actor = FlowActor(
                    provision=prov,
                    transport=self.transport,
                    flow=t,
                    is_generator=x.node.kind == NodeKind.GENERATOR,
                )

            except Exception as e:
                raise ProvisionException("No Actor Builder found for template") from e

        task = await actor.arun()
        self.provisionActorMap[prov.provision] = actor
        return actor
